#include <stdio.h>
#include <stdlib.h>

#define FIXNUM_MASK     3
#define FIXNUM_TAG      0
#define FIXNUM_SHIFT    2

#define CHAR_MASK       0xff
#define CHAR_SHIFT      8
#define CHAR_TAG        7

#define BOOL_MASK       0xff
#define BOOL_SHIFT      8
#define BOOL_TAG        15

#define PTR_MASK        7
#define PAIR_TAG        1
#define VEC_TAG         2
#define STR_TAG         3
#define SYM_TAG         5
#define CLOSURE_TAG     6

#define HEAP_SIZE       0x400000

void show(int x) {
    if((x & FIXNUM_MASK) == FIXNUM_TAG) {
        // integer
        printf("%d", x >> FIXNUM_SHIFT);
    } else if((x & CHAR_MASK) == CHAR_TAG) {
        // character
        printf("#\\%c", (char)(x >> CHAR_SHIFT));
    } else if((x & BOOL_MASK) == BOOL_TAG) {
        // boolean
        if((x >> BOOL_SHIFT) != 0) {
            printf("#t");
        } else {
            printf("#f");
        }
    } else if((x & PTR_MASK) == PAIR_TAG) {
        int* ptr = (int*)(x - PAIR_TAG);
        if(ptr == NULL) {
            printf("()");
            return;
        }

        // either a list or a dotted pair
        int car = ptr[0];
        int cdr = ptr[1];
        putchar('(');
        show(car);

        // show additional space-separated elems
        while((cdr & PTR_MASK) == PAIR_TAG) {
            ptr = (int*)(cdr - PAIR_TAG);
            if(ptr == NULL) break;

            car = ptr[0];
            cdr = ptr[1];
            putchar(' ');
            show(car);
        }

        // show dotted pair notation if relevant
        if((cdr & PTR_MASK) != PAIR_TAG) {
            printf(" . ");
            show(cdr);
        }
        putchar(')');
    } else if((x & PTR_MASK) == STR_TAG) {
        int* ptr = (int*)(x - STR_TAG);
        int len = *ptr;
        char* body = (char*)(ptr+1);
        putchar('"');
        for(;len > 0;len--) putchar(*body++);
        putchar('"');
    } else if((x & PTR_MASK) == VEC_TAG) {
        int* ptr = (int*)(x - VEC_TAG);
        int len = *ptr++;

        printf("#(");
        for(;len > 0;len--) {
            show(*ptr++);
            if(len != 1) putchar(' ');
        }
        printf(")");
    }
}

__attribute__((__cdecl__))
extern int scheme_entry(void* heap);

int main(int argc, const char **argv) {
    void* heap = aligned_alloc(8, HEAP_SIZE);
    if(heap == NULL) {
        printf("Memory allocation failed\n");
        return -1;
    }
    int val = scheme_entry(heap);
    show(val);
    printf("\n");
    free(heap);
    return 0;
}

